"""Commands for working with local profiles."""

__all__ = ["register"]

import sys

from . import (
    colorized,
    Command,
    print_with_pager,
    PROFILE_DEFAULT,
    PROFILE_NAMES,
    read_input,
    TableCommand,
    tables,
)
from .. import bones, utils
from ..bones import helpers
from ..utils import auth, profiles
from ..utils.maas_async import asynchronous


class cmd_login(Command):
    """Log-in to a MAAS with either username and password or apikey.

    The username and password will NOT be saved; a new API key will be
    obtained from MAAS and associated with the new profile. This key can be
    selectively revoked from the Web UI, for example, at a later date.
    """

    def __init__(self, parser):
        super(cmd_login, self).__init__(parser)
        parser.add_argument(
            "-p",
            "--profile-name",
            default=None,
            help=(
                "The name to give the profile. Default is the username used "
                "to login."
            ),
        )
        parser.add_argument(
            "--anonymous",
            default=False,
            action="store_true",
            help=(
                "Create an anonymous profile, no credentials are associated " "to it."
            ),
        )
        parser.add_argument(
            "--apikey",
            default=None,
            help=(
                "The API key acquired from MAAS. This requires the profile "
                "name to be provided as well."
            ),
        )
        parser.add_argument(
            "-k",
            "--insecure",
            action="store_true",
            help=("Disable SSL certificate check"),
            default=False,
        )
        parser.add_argument(
            "url",
            nargs="?",
            type=utils.api_url,
            help=(
                "The URL of the API, e.g. http://example.com/MAAS/ "
                "or http://example.com/MAAS/api/2.0/ if you wish to specify "
                "the API version. If no URL is provided then it will be "
                "prompted for, interactively."
            ),
        )
        parser.add_argument(
            "username",
            nargs="?",
            default=None,
            help=(
                "The username used to login to MAAS. If no username is "
                "provided and API key is not being used it will be prompted "
                "for, interactively."
            ),
        )
        parser.add_argument(
            "password",
            nargs="?",
            default=None,
            help=(
                "The password used to login to MAAS. If no password is "
                "proviced and API key is not being used it will be promoed "
                "for, interactively."
            ),
        )

    @asynchronous
    async def __call__(self, options):
        has_auth_info = any((options.apikey, options.username, options.password))
        if options.anonymous and has_auth_info:
            raise ValueError(
                "Can't specify username, password or--apikey with --anonymous"
            )

        if options.apikey and not options.profile_name:
            raise ValueError("-p,--profile-name must be provided with --apikey")

        if not options.url:
            url = read_input("URL", validator=utils.api_url)
        else:
            url = options.url

        if not options.apikey:
            if options.anonymous:
                password = None
            elif options.username and not options.password:
                password = read_input("Password", password=True)
            else:
                password = options.password
                if password == "-":
                    password = sys.stdin.readline().strip()
            try:
                profile = await helpers.login(
                    url,
                    anonymous=options.anonymous,
                    username=options.username,
                    password=password,
                    insecure=options.insecure,
                )
            except helpers.MacaroonLoginNotSupported:
                # the server doesn't have external authentication enabled,
                # propmt for username/password
                username = read_input("Username")
                password = read_input("Password", password=True)
                profile = await helpers.login(
                    url, username=username, password=password, insecure=options.insecure
                )
        else:
            credentials = auth.obtain_credentials(options.apikey)
            session = await bones.SessionAPI.fromURL(
                url, credentials=credentials, insecure=options.insecure
            )
            profile = profiles.Profile(
                options.profile_name,
                url,
                credentials=credentials,
                description=session.description,
            )

        if options.profile_name:
            profile = profile.replace(name=options.profile_name)

        # Save a new profile.
        with profiles.ProfileStore.open() as config:
            config.save(profile)
            config.default = profile

        self.print_whats_next(profile)

    @staticmethod
    def print_whats_next(profile):
        """Explain what to do next."""
        what_next = [
            "{{autogreen}}Congratulations!{{/autogreen}} You are logged in "
            "to the MAAS server at {{autoblue}}{profile.url}{{/autoblue}} "
            "with the profile name {{autoblue}}{profile.name}{{/autoblue}}.",
            "For help with the available commands, try:",
            "  maas help",
        ]
        for message in what_next:
            message = message.format(profile=profile)
            print(colorized(message))
            print()


class cmd_logout(Command):
    """Logout of a MAAS profile, purging any stored credentials.

    This will remove the given profile from your command-line client. You can
    re-create it later using `add` or `login`.
    """

    def __init__(self, parser):
        super(cmd_logout, self).__init__(parser)
        parser.add_argument(
            "profile_name",
            metavar="profile-name",
            nargs="?",
            choices=PROFILE_NAMES,
            help=(
                "The profile name you want to logout of."
                + ("" if PROFILE_DEFAULT is None else " [default: %(default)s]")
            ),
        )
        if PROFILE_DEFAULT is not None:
            parser.set_defaults(profile_name=PROFILE_DEFAULT.name)

    def __call__(self, options):
        with profiles.ProfileStore.open() as config:
            config.delete(options.profile_name)


class cmd_switch(Command):
    """Switch the active profile.

    This will switch the currently active profile to the given profile. The
    previous profile will remain, just use `switch` again to go back.
    """

    def __init__(self, parser):
        super(cmd_switch, self).__init__(parser)
        parser.add_argument(
            "profile_name",
            metavar="profile-name",
            choices=PROFILE_NAMES,
            help=("The profile name you want to switch to."),
        )

    def __call__(self, options):
        with profiles.ProfileStore.open() as config:
            profile = config.load(options.profile_name)
            config.default = profile


class cmd_profiles(TableCommand):
    """List profiles (aka. logged in MAAS's)."""

    def __init__(self, parser):
        super(cmd_profiles, self).__init__(parser)
        parser.add_argument(
            "--refresh",
            action="store_true",
            default=False,
            help=(
                "Retrieves the latest version of the help information for "
                "all profiles. Use it to update your command-line client's "
                "information after an upgrade to the MAAS server."
            ),
        )
        parser.other.add_argument(
            "--no-pager",
            action="store_true",
            help=("Don't use the pager when printing the output of the " "command."),
        )

    def __call__(self, options):
        if options.refresh:
            with profiles.ProfileStore.open() as config:
                for profile_name in config:
                    profile = config.load(profile_name)
                    session = bones.SessionAPI.fromProfile(profile)
                    profile = profile.replace(description=session.description)
                    config.save(profile)
        else:
            table = tables.ProfilesTable()
            with profiles.ProfileStore.open() as config:
                if options.no_pager:
                    print(table.render(options.format, config))
                else:
                    print_with_pager(table.render(options.format, config))


def register(parser):
    """Register profile commands with the given parser."""
    cmd_login.register(parser)
    cmd_logout.register(parser)
    cmd_switch.register(parser)
    cmd_profiles.register(parser)



"""Objects for region and rack controllers."""

__all__ = ["RackController", "RackControllers", "RegionController", "RegionControllers"]

from . import check, check_optional, ObjectField, to
from .nodes import Node, Nodes, NodesType, NodeTypeMeta
from ..enum import PowerState


class RackControllersType(NodesType):
    """Metaclass for `RackControllers`."""


class RackControllers(Nodes, metaclass=RackControllersType):
    """The set of rack-controllers stored in MAAS."""


class RackControllerType(NodeTypeMeta):
    """Metaclass for `RackController`."""


class RackController(Node, metaclass=RackControllerType):
    """A rack-controller stored in MAAS."""

    architecture = ObjectField.Checked(
        "architecture", check_optional(str), check_optional(str)
    )
    cpus = ObjectField.Checked("cpu_count", check(int), check(int))
    distro_series = ObjectField.Checked("distro_series", check(str), check(str))
    memory = ObjectField.Checked("memory", check(int), check(int))
    osystem = ObjectField.Checked("osystem", check(str), readonly=True)
    power_state = ObjectField.Checked("power_state", to(PowerState), readonly=True)

    # power_type
    # service_set
    # swap_size


class RegionControllersType(NodesType):
    """Metaclass for `RegionControllers`."""


class RegionControllers(Nodes, metaclass=RegionControllersType):
    """The set of region-controllers stored in MAAS."""


class RegionControllerType(NodeTypeMeta):
    """Metaclass for `RegionController`."""


class RegionController(Node, metaclass=RegionControllerType):
    """A region-controller stored in MAAS."""

    architecture = ObjectField.Checked(
        "architecture", check_optional(str), check_optional(str)
    )
    cpus = ObjectField.Checked("cpu_count", check(int), check(int))
    distro_series = ObjectField.Checked("distro_series", check(str), check(str))
    memory = ObjectField.Checked("memory", check(int), check(int))
    osystem = ObjectField.Checked("osystem", check(str), readonly=True)
    power_state = ObjectField.Checked("power_state", to(PowerState), readonly=True)

    # power_type
    # service_set
    # swap_size



"""Objects for vlans."""

__all__ = ["Vlan", "Vlans"]

from operator import attrgetter
from typing import Union

from . import (
    check,
    check_optional,
    Object,
    ObjectField,
    ObjectFieldRelated,
    ObjectSet,
    ObjectType,
)
from .controllers import RackController
from .fabrics import Fabric
from .spaces import Space


class VlanType(ObjectType):
    """Metaclass for `Vlan`."""

    async def read(cls, fabric: Union[Fabric, int], vid: int):
        """Get `Vlan` by `vid`.

        :param fabric: Fabric to get the VLAN from.
        :type fabric: `Fabric` or `int`
        :param vid: VID of VLAN.
        :type vid: `int`
        """
        if isinstance(fabric, int):
            fabric_id = fabric
        elif isinstance(fabric, Fabric):
            fabric_id = fabric.id
        else:
            raise TypeError(
                "fabric must be a Fabric or int, not %s" % type(fabric).__name__
            )
        data = await cls._handler.read(fabric_id=fabric_id, vid=vid)
        return cls(data, {"fabric_id": fabric_id})


class Vlan(Object, metaclass=VlanType):
    """A VLAN in a fabric."""

    id = ObjectField.Checked("id", check(int), readonly=True)

    fabric = ObjectFieldRelated("fabric_id", "Fabric", pk=0)
    vid = ObjectField.Checked("vid", check(int), check(int), pk=1)

    name = ObjectField.Checked("name", check_optional(str), check_optional(str))
    mtu = ObjectField.Checked("mtu", check(int), check(int))
    space = ObjectFieldRelated(
        "space",
        "Space",
        map_func=lambda _, data: {"name": data, "__incomplete__": True},
    )

    relay_vlan = ObjectFieldRelated("relay_vlan", "Vlan", use_data_setter=True)
    dhcp_on = ObjectField.Checked("dhcp_on", check(bool), check(bool))
    primary_rack = ObjectFieldRelated("primary_rack", "RackController")
    secondary_rack = ObjectFieldRelated("secondary_rack", "RackController")

    external_dhcp = ObjectField.Checked(
        "external_dhcp", check_optional(str), readonly=True
    )

    def __repr__(self):
        return super(Vlan, self).__repr__(fields={"name", "vid"})

    async def save(self):
        # Store fabric_id so its reset on save. `fabric_id` is not returned
        # by the API, this is stored through `local_data`.
        fabric_id = self._data["fabric_id"]
        if (
            "fabric_id" in self._changed_data
            and self._changed_data["fabric_id"] is not None
        ):
            if self._orig_data["fabric_id"] and self._changed_data["fabric_id"] == (
                self._orig_data["fabric_id"]
            ):
                # Fabric didn't really change, the object was just set to the
                # same Fabric
                del self._changed_data["fabric_id"]
            else:
                # Allow the update of the fabric by setting the 'fabric' field
                # with the new 'fabric_id'.
                self._changed_data["fabric"] = self._changed_data["fabric_id"]

        if "relay_vlan" in self._changed_data and self._changed_data["relay_vlan"]:
            # Update uses the ID of the VLAN, not the VLAN object.
            self._changed_data["relay_vlan"] = self._changed_data["relay_vlan"]["id"]
            if (
                "relay_vlay" in self._orig_data
                and self._orig_data["relay_vlan"]
                and "id" in self._orig_data["relay_vlan"]
                and self._changed_data["relay_vlan"]
                == (self._orig_data["relay_vlan"]["id"])
            ):
                # VLAN didn't really change, the object was just set to the
                # same VLAN.
                del self._changed_data["relay_vlan"]
            else:
                self._changed_data["relay_vlan"] = self._changed_data["relay_vlan"]

        # Don't call `super().save()` because `vid` can be changed and it has
        # to be handled specially because its in the resource_uri path.
        if self._changed_data:
            update_data = dict(self._changed_data)
            update_data.update(
                {key: self._orig_data[key] for key in self._handler.params}
            )
            update_data["vid"] = update_data["_vid"] = self._orig_data["vid"]
            if "vid" in self._changed_data:
                update_data["_vid"] = self._changed_data["vid"]
            data = await self._handler.update(**update_data)
            # Set fabric_id because it was lost from the `update`.
            data["fabric_id"] = fabric_id
            self._reset(data)

    async def delete(self):
        """Delete this VLAN."""
        # Since the VID can be changed for the VLAN, we always use the vid
        # from the original data. That way if the user changes the vid the
        # delete still works, until the VLAN has been saved.
        await self._handler.delete(fabric_id=self.fabric.id, vid=self._orig_data["vid"])


class VlansType(ObjectType):
    """Metaclass for `Vlans`."""

    async def read(cls, fabric: Union[Fabric, int]):
        """Get list of `Vlan`'s for `fabric`.

        :param fabric: Fabric to get all VLAN's for.
        :type fabric: `Fabric` or `int`
        """
        if isinstance(fabric, int):
            fabric_id = fabric
        elif isinstance(fabric, Fabric):
            fabric_id = fabric.id
        else:
            raise TypeError(
                "fabric must be a Fabric or int, not %s" % type(fabric).__name__
            )
        data = await cls._handler.read(fabric_id=fabric_id)
        return cls(
            cls._object(item, local_data={"fabric_id": fabric_id}) for item in data
        )

    async def create(
        cls,
        fabric: Union[Fabric, int],
        vid: int,
        *,
        name: str = None,
        description: str = None,
        mtu: int = None,
        relay_vlan: Union[Vlan, int] = None,
        dhcp_on: bool = False,
        primary_rack: Union[RackController, str] = None,
        secondary_rack: Union[RackController, str] = None,
        space: Union[Space, int] = None
    ):
        """
        Create a `Vlan` in MAAS.

        :param fabric: Fabric to create the VLAN on.
        :type fabric: `Fabric` or `int`
        :param vid: VID for the VLAN.
        :type vid: `int`
        :param name: The name of the VLAN (optional).
        :type name: `str`
        :param description: A description of the VLAN (optional).
        :type description: `str`
        :param mtu: The MTU for VLAN (optional, default of 1500 will be used).
        :type mtu: `int`
        :param relay_vlan: VLAN to relay this VLAN through.
        :type relay_vlan: `Vlan` or `int`
        :param dhcp_on: True turns the DHCP on, false keeps the DHCP off. True
            requires that `primary_rack` is also set.
        :type dhcp_on: `bool`
        :param primary_rack: Primary rack controller to run the DCHP
            service on.
        :type primary_rack: `RackController` or `int`
        :parma secondary_rack: Secondary rack controller to run the DHCP
            service on. This will enable HA operation of the DHCP service.
        :type secondary_rack: `RackController` or `int`
        :returns: The created VLAN.
        :rtype: `Vlan`
        """
        params = {}
        if isinstance(fabric, int):
            params["fabric_id"] = fabric
        elif isinstance(fabric, Fabric):
            params["fabric_id"] = fabric.id
        else:
            raise TypeError(
                "fabric must be Fabric or int, not %s" % (type(fabric).__class__)
            )
        params["vid"] = vid
        if name is not None:
            params["name"] = name
        if description is not None:
            params["description"] = description
        if mtu is not None:
            params["mtu"] = mtu
        if relay_vlan is not None:
            if isinstance(relay_vlan, int):
                params["relay_vlan"] = relay_vlan
            elif isinstance(relay_vlan, Vlan):
                params["relay_vlan"] = relay_vlan.id
            else:
                raise TypeError(
                    "relay_vlan must be Vlan or int, not %s"
                    % (type(relay_vlan).__class__)
                )
        params["dhcp_on"] = dhcp_on
        if primary_rack is not None:
            if isinstance(primary_rack, str):
                params["primary_rack"] = primary_rack
            elif isinstance(primary_rack, RackController):
                params["primary_rack"] = primary_rack.system_id
            else:
                raise TypeError(
                    "primary_rack must be RackController or str, not %s"
                    % (type(primary_rack).__class__)
                )
        if secondary_rack is not None:
            if isinstance(secondary_rack, str):
                params["secondary_rack"] = secondary_rack
            elif isinstance(secondary_rack, RackController):
                params["secondary_rack"] = secondary_rack.system_id
            else:
                raise TypeError(
                    "secondary_rack must be RackController or str, not %s"
                    % (type(secondary_rack).__class__)
                )
        if space is not None:
            if isinstance(space, int):
                params["space"] = space
            elif isinstance(space, Space):
                params["space"] = space.id
            else:
                raise TypeError(
                    "space must be Space or int, not %s" % (type(space).__class__)
                )
        return cls._object(await cls._handler.create(**params))


class Vlans(ObjectSet, metaclass=VlansType):
    """The set of VLAN's on a fabric."""

    def get_default(self):
        """Return the default VLAN from the set."""
        length = len(self)
        if length == 0:
            return None
        elif length == 1:
            return self[0]
        else:
            return sorted(self, key=attrgetter("id"))[0]



"""Objects for pods."""

__all__ = ["Pod", "Pods"]

import typing
from . import check, Object, ObjectField, ObjectFieldRelated, ObjectSet, ObjectType
from .zones import Zone
from ..utils import remove_None
from ..errors import OperationNotAllowed


class PodsType(ObjectType):
    """Metaclass for `Pods`."""

    async def read(cls):
        data = await cls._handler.read()
        return cls(map(cls._object, data))

    async def create(
        cls,
        *,
        type: str,
        power_address: str,
        power_user: str = None,
        power_pass: str = None,
        name: str = None,
        zone: typing.Union[str, Zone] = None,
        tags: typing.Sequence[str] = None
    ):
        """Create a `Pod` in MAAS.

        :param type: Type of pod to create (rsd, virsh) (required).
        :type name: `str`
        :param power_address: Address for power control of the pod (required).
        :type power_address: `str`
        :param power_user: User for power control of the pod
            (required for rsd).
        :type power_user: `str`
        :param power_pass: Password for power control of the pod
            (required for rsd).
        :type power_pass: `str`
        :param name: Name for the pod (optional).
        :type name: `str`
        :param zone: Name of the zone for the pod (optional).
        :type zone: `str` or `Zone`
        :param tags: A tag or tags (separated by comma) for the pod.
        :type tags: `str`
        :returns: The created Pod.
        :rtype: `Pod`
        """
        params = remove_None(
            {
                "type": type,
                "power_address": power_address,
                "power_user": power_user,
                "power_pass": power_pass,
                "name": name,
                "tags": tags,
            }
        )
        if type == "rsd" and power_user is None:
            message = "'power_user' is required for pod type `rsd`"
            raise OperationNotAllowed(message)
        if type == "rsd" and power_pass is None:
            message = "'power_pass' is required for pod type `rsd`"
            raise OperationNotAllowed(message)
        if zone is not None:
            if isinstance(zone, Zone):
                params["zone"] = zone.name
            elif isinstance(zone, str):
                params["zone"] = zone
            else:
                raise TypeError(
                    "zone must be a str or Zone, not %s" % type(zone).__name__
                )
        return cls._object(await cls._handler.create(**params))


class Pods(ObjectSet, metaclass=PodsType):
    """The set of `Pods` stored in MAAS."""


class PodType(ObjectType):
    """Metaclass for a `Pod`."""

    async def read(cls, id: int):
        """Get `Pod` by `id`."""
        data = await cls._handler.read(id=id)
        return cls(data)


class Pod(Object, metaclass=PodType):
    """A `Pod` stored in MAAS."""

    id = ObjectField.Checked("id", check(int), readonly=True, pk=True)
    type = ObjectField.Checked("type", check(str), check(str))
    name = ObjectField.Checked("name", check(str), check(str))
    architectures = ObjectField.Checked("architectures", check(list), check(list))
    capabilities = ObjectField.Checked("capabilities", check(list), check(list))
    zone = ObjectFieldRelated("zone", "Zone", readonly=True)
    tags = ObjectField.Checked("tags", check(list), check(list))
    cpu_over_commit_ratio = ObjectField.Checked(
        "cpu_over_commit_ratio", check(float), check(float)
    )
    memory_over_commit_ratio = ObjectField.Checked(
        "memory_over_commit_ratio", check(float), check(float)
    )
    available = ObjectField.Checked("available", check(dict), check(dict))
    used = ObjectField.Checked("used", check(dict), check(dict))
    total = ObjectField.Checked("total", check(dict), check(dict))
    default_macvlan_mode = ObjectField.Checked(
        "default_macvlan_mode", check(str), check(str)
    )
    host = ObjectFieldRelated("host", "Node", readonly=True)

    async def save(self):
        """Save the `Pod`."""
        old_tags = list(self._orig_data["tags"])
        new_tags = list(self.tags)
        self._changed_data.pop("tags", None)
        await super(Pod, self).save()
        for tag_name in new_tags:
            if tag_name not in old_tags:
                await self._handler.add_tag(id=self.id, tag=tag_name)
            else:
                old_tags.remove(tag_name)
        for tag_name in old_tags:
            await self._handler.remove_tag(id=self.id, tag=tag_name)
        self._orig_data["tags"] = new_tags
        self._data["tags"] = list(new_tags)

    async def refresh(self):
        """Refresh the `Pod`."""
        return await self._handler.refresh(id=self.id)

    async def parameters(self):
        """Get the power parameters for the `Pod`."""
        return await self._handler.parameters(id=self.id)

    async def compose(
        self,
        *,
        cores: int = None,
        memory: int = None,
        cpu_speed: int = None,
        architecture: str = None,
        storage: typing.Sequence[str] = None,
        hostname: str = None,
        domain: typing.Union[int, str] = None,
        zone: typing.Union[int, str, Zone] = None,
        interfaces: typing.Sequence[str] = None
    ):
        """Compose a machine from `Pod`.

        All fields below are optional:

        :param cores: Minimum number of CPU cores.
        :type cores: `int`
        :param memory: Minimum amount of memory (MiB).
        :type memory: `int`
        :param cpu_speed: Minimum amount of CPU speed (MHz).
        :type cpu_speed: `int`
        :param architecture: Architecture for the machine. Must be an
            architecture that the pod supports.
        :type architecture: `str`
        :param storage: A list of storage constraint identifiers, in the form:
            <label>:<size>(<tag>[,<tag>[,...])][,<label>:...]
        :type storage: `str`
        :param hostname: Hostname for the newly composed machine.
        :type hostname: `str`
        :param domain: The domain ID for the machine (optional).
        :type domain: `int` or `str`
        :param zone: The zone ID for the machine (optional).
        :type zone: `int` or 'str' or `Zone`
        :param interfaces: A labeled constraint map associating constraint
            labels with interface properties that should be matched. Returned
            nodes must have one or more interface matching the specified
            constraints. The labeled constraint map must be in the format:
            ``<label>:<key>=<value>[,<key2>=<value2>[,...]]``

            Each key can be one of the following:

            - id: Matches an interface with the specific id
            - fabric: Matches an interface attached to the specified fabric.
            - fabric_class: Matches an interface attached to a fabric
              with the specified class.
            - ip: Matches an interface with the specified IP address
              assigned to it.
            - mode: Matches an interface with the specified mode. (Currently,
              the only supported mode is "unconfigured".)
            - name: Matches an interface with the specified name.
              (For example, "eth0".)
            - hostname: Matches an interface attached to the node with
              the specified hostname.
            - subnet: Matches an interface attached to the specified subnet.
            - space: Matches an interface attached to the specified space.
            - subnet_cidr: Matches an interface attached to the specified
              subnet CIDR. (For example, "192.168.0.0/24".)
            - type: Matches an interface of the specified type. (Valid
              types: "physical", "vlan", "bond", "bridge", or "unknown".)
            - vlan: Matches an interface on the specified VLAN.
            - vid: Matches an interface on a VLAN with the specified VID.
            - tag: Matches an interface tagged with the specified tag.
        :type interfaces: `str`
        :returns: The created Machine
        :rtype: `Machine`
        """
        params = remove_None(
            {
                "cores": str(cores) if cores else None,
                "memory": str(memory) if memory else None,
                "cpu_speed": str(cpu_speed) if cpu_speed else None,
                "architecture": architecture,
                "storage": storage,
                "hostname": hostname,
                "domain": str(domain) if domain else None,
                "interfaces": interfaces,
            }
        )
        if zone is not None:
            if isinstance(zone, Zone):
                params["zone"] = str(zone.id)
            elif isinstance(zone, int):
                params["zone"] = str(zone)
            elif isinstance(zone, str):
                params["zone"] = zone
            else:
                raise TypeError(
                    "zone must be an int, str or Zone, not %s" % type(zone).__name__
                )
        return await self._handler.compose(**params, id=self.id)

    async def delete(self):
        """Delete the `Pod`."""
        return await self._handler.delete(id=self.id)



"""Tables for representing information from MAAS."""

__all__ = ["FilesTable", "NodesTable", "ProfilesTable", "TagsTable", "UsersTable"]

from operator import attrgetter, itemgetter

from colorclass import Color

from ..enum import InterfaceType, NodeType, PowerState, RDNSMode
from .tabular import Column, RenderTarget, Table, DetailTable, NestedTableColumn


class NodeTypeColumn(Column):

    nice_names = {
        NodeType.MACHINE: "Machine",
        NodeType.DEVICE: "Device",
        NodeType.RACK_CONTROLLER: "Rackd",
        NodeType.REGION_CONTROLLER: "Regiond",
        NodeType.REGION_AND_RACK_CONTROLLER: "Regiond+rackd",
    }

    def render(self, target, node_type):
        if target in (RenderTarget.pretty, RenderTarget.plain):
            node_type = self.nice_names[node_type]
        else:
            node_type = node_type.value
        return super().render(target, node_type)


class NodeArchitectureColumn(Column):
    def render(self, target, architecture):
        if target in (RenderTarget.pretty, RenderTarget.plain):
            if architecture:
                if architecture.endswith("/generic"):
                    architecture = architecture[:-8]
            else:
                architecture = "-"
        return super().render(target, architecture)


class NodeCPUsColumn(Column):
    def render(self, target, cpus):
        # `cpus` is a count of CPUs.
        if target in (RenderTarget.pretty, RenderTarget.plain):
            if cpus is None:
                cpus = "-"
            elif cpus == 0.0:
                cpus = "-"
        return super().render(target, cpus)


class NodeMemoryColumn(Column):
    def render(self, target, memory):
        # `memory` is in MB.
        if target in (RenderTarget.pretty, RenderTarget.plain):
            if memory is None:
                memory = "-"
            elif memory == 0.0:
                memory = "-"
            elif memory < 1024.0:  # <1GB
                memory = "%0.1f MB" % (memory)
            elif memory < 1048576.0:  # <1TB
                memory = "%0.1f GB" % (memory / 1024.0)
            else:
                memory = "%0.1f TB" % (memory / 1048576.0)
        return super().render(target, memory)


class NodeStatusNameColumn(Column):

    colours = {
        # "New": "",  # White.
        "Commissioning": "autoyellow",
        "Failed commissioning": "autored",
        "Missing": "red",
        "Ready": "autogreen",
        "Reserved": "autoblue",
        "Allocated": "autoblue",
        "Deploying": "autoblue",
        "Deployed": "autoblue",
        # "Retired": "",  # White.
        "Broken": "autored",
        "Failed deployment": "autored",
        "Releasing": "autoblue",
        "Releasing failed": "autored",
        "Disk erasing": "autoblue",
        "Failed disk erasing": "autored",
    }

    def render(self, target, datum):
        if target == RenderTarget.pretty:
            if datum in self.colours:
                colour = self.colours[datum]
                return Color("{%s}%s{/%s}" % (colour, datum, colour))
            else:
                return datum
        else:
            return super().render(target, datum)


class NodePowerColumn(Column):

    colours = {
        PowerState.ON: "autogreen",
        # PowerState.OFF: "",  # White.
        PowerState.ERROR: "autored",
    }

    def render(self, target, data):
        if target == RenderTarget.pretty:
            if data in self.colours:
                colour = self.colours[data]
                return Color("{%s}%s{/%s}" % (colour, data.value.capitalize(), colour))
            else:
                return data.value.capitalize()
        elif target == RenderTarget.plain:
            return super().render(target, data.value.capitalize())
        else:
            return super().render(target, data.value)


class NodeInterfacesColumn(Column):
    def render(self, target, data):
        count = 0
        for interface in data:
            if interface.type == InterfaceType.PHYSICAL:
                count += 1
        return super().render(target, "%d physical" % count)


class NodeOwnerColumn(Column):
    def render(self, target, data):
        if data is None:
            return super().render(target, "(none)")
        else:
            return super().render(target, data.username)


class NodeImageColumn(Column):
    def render(self, target, data):
        if data:
            return super().render(target, data)
        else:
            return super().render(target, "(none)")


class NodeZoneColumn(Column):
    def render(self, target, data):
        return super().render(target, data.name)


class NodeResourcePoolColumn(Column):
    def render(self, target, data):
        return super().render(target, data.name)


class NodeTagsColumn(Column):
    def render(self, target, data):
        if data:
            return super().render(target, [tag.name for tag in data])
        else:
            return ""


class NodesTable(Table):
    def __init__(self):
        super().__init__(
            Column("hostname", "Hostname"), NodeTypeColumn("node_type", "Type")
        )

    def get_rows(self, target, nodes):
        data = ((node.hostname, node.node_type) for node in nodes)
        return sorted(data, key=itemgetter(0))


class MachinesTable(Table):
    def __init__(self):
        super().__init__(
            Column("hostname", "Hostname"),
            NodePowerColumn("power", "Power"),
            NodeStatusNameColumn("status", "Status"),
            NodeOwnerColumn("owner", "Owner"),
            NodeArchitectureColumn("architecture", "Arch"),
            NodeCPUsColumn("cpus", "#CPUs"),
            NodeMemoryColumn("memory", "RAM"),
            NodeResourcePoolColumn("pool", "Resource pool"),
            NodeZoneColumn("zone", "Zone"),
        )

    def get_rows(self, target, machines):
        data = (
            (
                machine.hostname,
                machine.power_state,
                machine.status_name,
                machine.owner,
                machine.architecture,
                machine.cpus,
                machine.memory,
                machine.pool,
                machine.zone,
            )
            for machine in machines
        )
        return sorted(data, key=itemgetter(0))


class MachineDetail(DetailTable):
    def __init__(self, with_type=False):
        self.with_type = with_type
        columns = [
            Column("hostname", "Hostname"),
            NodeStatusNameColumn("status", "Status"),
            NodeImageColumn("image", "Image"),
            NodePowerColumn("power", "Power"),
            Column("power_type", "Power Type"),
            NodeArchitectureColumn("architecture", "Arch"),
            NodeCPUsColumn("cpus", "#CPUs"),
            NodeMemoryColumn("memory", "RAM"),
            NodeInterfacesColumn("interfaces", "Interfaces"),
            Column("ip_addresses", "IP addresses"),
            NodeResourcePoolColumn("pool", "Resource pool"),
            NodeZoneColumn("zone", "Zone"),
            NodeOwnerColumn("owner", "Owner"),
            NodeTagsColumn("tags", "Tags"),
        ]
        if with_type:
            columns.insert(1, NodeTypeColumn("node_type", "Type"))
        super().__init__(*columns)

    def get_rows(self, target, machine):
        data = [
            machine.hostname,
            machine.status_name,
            machine.distro_series,
            machine.power_state,
            machine.power_type,
            machine.architecture,
            machine.cpus,
            machine.memory,
            machine.interfaces,
            [
                link.ip_address
                for interface in machine.interfaces
                for link in interface.links
                if link.ip_address
            ],
            machine.pool,
            machine.zone,
            machine.owner,
            machine.tags,
        ]
        if self.with_type:
            data.insert(1, machine.node_type)
        return data


class DevicesTable(Table):
    def __init__(self):
        super().__init__(
            Column("hostname", "Hostname"),
            NodeOwnerColumn("owner", "Owner"),
            Column("ip_addresses", "IP addresses"),
        )

    def get_rows(self, target, devices):
        data = (
            (
                device.hostname,
                device.owner,
                [
                    link.ip_address
                    for interface in device.interfaces
                    for link in interface.links
                    if link.ip_address
                ],
            )
            for device in devices
        )
        return sorted(data, key=itemgetter(0))


class DeviceDetail(DetailTable):
    def __init__(self, with_type=False):
        self.with_type = with_type
        columns = [
            Column("hostname", "Hostname"),
            NodeInterfacesColumn("interfaces", "Interfaces"),
            Column("ip_addresses", "IP addresses"),
            NodeOwnerColumn("owner", "Owner"),
            Column("tags", "Tags"),
        ]
        if with_type:
            columns.insert(1, NodeTypeColumn("node_type", "Type"))
        super().__init__(*columns)

    def get_rows(self, target, device):
        data = [
            device.hostname,
            device.interfaces,
            [
                link.ip_address
                for interface in device.interfaces
                for link in interface.links
                if link.ip_address
            ],
            device.owner,
            device.tags,
        ]
        if self.with_type:
            data.insert(1, device.node_type)
        return data


class ControllersTable(Table):
    def __init__(self):
        super().__init__(
            Column("hostname", "Hostname"),
            NodeTypeColumn("node_type", "Type"),
            NodeArchitectureColumn("architecture", "Arch"),
            NodeCPUsColumn("cpus", "#CPUs"),
            NodeMemoryColumn("memory", "RAM"),
        )

    def get_rows(self, target, controllers):
        data = (
            (
                controller.hostname,
                controller.node_type,
                controller.architecture,
                controller.cpus,
                controller.memory,
            )
            for controller in controllers
        )
        return sorted(data, key=itemgetter(0))


class ControllerDetail(DetailTable):
    def __init__(self):
        super().__init__(
            Column("hostname", "Hostname"),
            NodeTypeColumn("node_type", "Type"),
            NodeArchitectureColumn("architecture", "Arch"),
            NodeCPUsColumn("cpus", "#CPUs"),
            NodeMemoryColumn("memory", "RAM"),
            NodeInterfacesColumn("interfaces", "Interfaces"),
            NodeZoneColumn("zone", "Zone"),
        )

    def get_rows(self, target, controller):
        return (
            controller.hostname,
            controller.node_type,
            controller.architecture,
            controller.cpus,
            controller.memory,
            controller.interfaces,
            controller.zone,
        )


class TagsTable(Table):
    def __init__(self):
        super().__init__(
            Column("name", "Tag name"),
            Column("definition", "Definition"),
            Column("kernel_opts", "Kernel options"),
            Column("comment", "Comment"),
        )

    def get_rows(self, target, tags):
        data = (
            (tag.name, tag.definition, tag.kernel_opts, tag.comment) for tag in tags
        )
        return sorted(data, key=itemgetter(0))


class FilesTable(Table):
    def __init__(self):
        super().__init__(Column("filename", "File name"))

    def get_rows(self, target, files):
        data = ((f.filename,) for f in files)
        return sorted(data, key=itemgetter(0))


class UserIsAdminColumn(Column):

    yes, no = "Yes", "Np"
    yes_pretty = Color("{autogreen}Yes{/autogreen}")

    def render(self, target, is_admin):
        if target is RenderTarget.plain:
            return self.yes if is_admin else self.no
        elif target is RenderTarget.pretty:
            return self.yes_pretty if is_admin else self.no
        else:
            return super().render(target, is_admin)


class UsersTable(Table):
    def __init__(self):
        super().__init__(
            Column("username", "User name"),
            Column("email", "Email address"),
            UserIsAdminColumn("is_admin", "Admin?"),
        )

    def get_rows(self, target, users):
        data = ((user.username,) for user in users)
        return sorted(data, key=itemgetter(0))


class ProfileActiveColumn(Column):
    def render(self, target, is_anonymous):
        if target is RenderTarget.pretty:
            return "✓" if is_anonymous else " "
        elif target is RenderTarget.plain:
            return "X" if is_anonymous else " "
        else:
            return super().render(target, is_anonymous)


class ProfilesTable(Table):
    def __init__(self):
        super().__init__(
            Column("name", "Profile"),
            Column("url", "URL"),
            ProfileActiveColumn("is_default", "Active"),
        )

    def get_rows(self, target, profiles):
        default = profiles.default
        default_name = None if default is None else default.name
        data = (
            (profile.name, profile.url, (profile.name == default_name))
            for profile in (profiles.load(name) for name in profiles)
        )
        return sorted(data, key=itemgetter(0))


class VIDColumn(Column):
    def render(self, target, data):
        vlan, vlans = data, data._data["fabric"].vlans
        vlans = sorted(vlans, key=attrgetter("id"))
        if vlans[0] == vlan:
            if vlan.vid == 0:
                data = "untagged"
            else:
                data = "untagged (%d)" % vlan.vid
        else:
            data = vlan.vid
        return super().render(target, data)


class DHCPColumn(Column):
    def render(self, target, vlan):
        if vlan.dhcp_on:
            if vlan.primary_rack:
                if vlan.secondary_rack:
                    text = "HA Enabled"
                else:
                    text = "Enabled"
            if target == RenderTarget.pretty:
                text = Color("{autogreen}%s{/autogreen}") % text
        elif vlan.relay_vlan:
            text = "Relayed via %s.%s" % (vlan.fabric.name, vlan.vid)
            if target == RenderTarget.pretty:
                text = Color("{autoblue}%s{/autoblue}") % text
        else:
            text = "Disabled"
        return super().render(target, text)


class SpaceNameColumn(Column):
    def render(self, target, space):
        name = space.name
        if name == "undefined":
            name = "(undefined)"
        return super().render(target, name)


class SubnetNameColumn(Column):
    def render(self, target, subnet):
        name = subnet.cidr
        if subnet.name and subnet.name != name:
            name = "%s (%s)" % (name, subnet.name)
        return super().render(target, name)


class SubnetActiveColumn(Column):
    def render(self, target, active):
        if active:
            text = "Active"
        else:
            text = "Disabled"
        if target == RenderTarget.pretty and active:
            text = Color("{autogreen}Active{/autogreen}")
        return super().render(target, text)


class SubnetRDNSModeColumn(Column):
    def render(self, target, mode):
        if mode == RDNSMode.DISABLED:
            text = "Disabled"
        else:
            if mode == RDNSMode.ENABLED:
                text = "Enabled"
            elif mode == RDNSMode.RFC2317:
                text = "Enabled w/ RFC 2317"
            if target == RenderTarget.pretty:
                text = Color("{autogreen}%s{/autogreen}") % text
        return super().render(target, text)


class SubnetsTable(Table):
    def __init__(self, *, visible_columns=None, fabrics=None):
        self.fabrics = fabrics
        super().__init__(
            SubnetNameColumn("name", "Subnet"),
            VIDColumn("vid", "VID"),
            Column("fabric", "Fabric"),
            SpaceNameColumn("space", "Space"),
            visible_columns=visible_columns,
        )

    def get_vlan(self, vlan):
        fabric = self.get_fabric(vlan.fabric)
        for fabric_vlan in fabric.vlans:
            if fabric_vlan.id == vlan.id:
                return fabric_vlan

    def get_fabric(self, unloaded_fabric):
        for fabric in self.fabrics:
            if fabric.id == unloaded_fabric.id:
                return fabric

    def get_rows(self, target, subnets):
        return (
            (
                subnet,
                self.get_vlan(subnet.vlan),
                self.get_fabric(subnet.vlan.fabric).name,
                subnet.vlan.space,
            )
            for subnet in sorted(subnets, key=attrgetter("cidr"))
        )


class SubnetDetail(DetailTable):
    def __init__(self, *, fabrics=None):
        self.fabrics = fabrics
        super().__init__(
            Column("name", "Name"),
            Column("cidr", "CIDR"),
            Column("gateway_ip", "Gateway IP"),
            Column("dns", "DNS"),
            VIDColumn("vid", "VID"),
            Column("fabric", "Fabric"),
            SpaceNameColumn("space", "Space"),
            SubnetActiveColumn("managed", "Managed allocation"),
            SubnetActiveColumn("allow_proxy", "Allow proxy"),
            SubnetActiveColumn("active_discovery", "Active discovery"),
            SubnetRDNSModeColumn("rdns_mode", "Reverse DNS mode"),
        )

    def get_vlan(self, vlan):
        fabric = self.get_fabric(vlan.fabric)
        for fabric_vlan in fabric.vlans:
            if fabric_vlan.id == vlan.id:
                return fabric_vlan

    def get_fabric(self, unloaded_fabric):
        for fabric in self.fabrics:
            if fabric.id == unloaded_fabric.id:
                return fabric

    def get_rows(self, target, subnet):
        return (
            subnet.name,
            subnet.cidr,
            subnet.gateway_ip,
            subnet.dns_servers,
            self.get_vlan(subnet.vlan),
            self.get_fabric(subnet.vlan.fabric).name,
            subnet.vlan.space,
            subnet.managed,
            subnet.allow_proxy,
            subnet.active_discovery,
            subnet.rdns_mode,
        )


class VlansTable(Table):
    def __init__(self, *, visible_columns=None, fabrics=None, subnets=None):
        self.subnets = subnets
        super().__init__(
            VIDColumn("vid", "VID"),
            DHCPColumn("dhcp", "DHCP"),
            SpaceNameColumn("space", "Space"),
            NestedTableColumn(
                "subnets",
                "Subnets",
                SubnetsTable,
                None,
                {"visible_columns": ("name",), "fabrics": fabrics},
            ),
            visible_columns=visible_columns,
        )

    def get_subnets(self, vlan):
        """Return the subnets for the `vlan`."""
        return vlan._origin.Subnets(
            [subnet for subnet in self.subnets if subnet.vlan.id == vlan.id]
        )

    def get_rows(self, target, vlans):
        return (
            (vlan, vlan, vlan.space, self.get_subnets(vlan))
            for vlan in sorted(vlans, key=attrgetter("vid"))
        )


class VlanDetail(DetailTable):
    def __init__(self, *, fabrics=None, subnets=None):
        self.fabrics = fabrics
        self.subnets = subnets
        super().__init__(
            VIDColumn("vid", "VID"),
            Column("name", "Name"),
            Column("fabric", "Fabric"),
            Column("mtu", "MTU"),
            DHCPColumn("dhcp", "DHCP"),
            Column("primary_rack", "Primary rack"),
            Column("secondary_rack", "Secondary rack"),
            SpaceNameColumn("space", "Space"),
            NestedTableColumn(
                "subnets",
                "Subnets",
                SubnetsTable,
                None,
                {"visible_columns": ("name",), "fabrics": fabrics},
            ),
        )

    def get_fabric(self, vlan):
        for fabric in self.fabrics:
            if fabric.id == vlan.fabric.id:
                return fabric

    def get_subnets(self, vlan):
        """Return the subnets for the `vlan`."""
        return vlan._origin.Subnets(
            [subnet for subnet in self.subnets if subnet.vlan.id == vlan.id]
        )

    def get_rows(self, target, vlan):
        primary_rack = vlan.primary_rack
        if primary_rack is not None:
            primary_rack.refresh()
        secondary_rack = vlan.secondary_rack
        if secondary_rack is not None:
            secondary_rack.refresh()
        return (
            vlan,
            vlan.name,
            self.get_fabric(vlan).name,
            vlan.mtu,
            vlan,
            primary_rack.hostname if primary_rack else None,
            secondary_rack.hostname if secondary_rack else None,
            vlan.space,
            self.get_subnets(vlan),
        )


class FabricsTable(Table):
    def __init__(self, *, visible_columns=None, subnets=None):
        super().__init__(
            Column("name", "Fabric"),
            NestedTableColumn("vlans", "VLANs", VlansTable, None, {"subnets": subnets}),
            visible_columns=visible_columns,
        )

    def get_rows(self, target, fabrics):
        self["vlans"].table_kwargs["fabrics"] = fabrics
        return (
            (fabric.name, fabric.vlans)
            for fabric in sorted(fabrics, key=attrgetter("id"))
        )


class FabricDetail(DetailTable):
    def __init__(self, *, fabrics=None, subnets=None):
        super().__init__(
            Column("name", "Name"),
            NestedTableColumn(
                "vlans",
                "VLANs",
                VlansTable,
                None,
                {"fabrics": fabrics, "subnets": subnets},
            ),
        )

    def get_rows(self, target, fabric):
        return (fabric.name, fabric.vlans)


class SpacesTable(Table):
    def __init__(self, *, visible_columns=None, fabrics=None, subnets=None):
        self.fabrics = fabrics
        super().__init__(
            SpaceNameColumn("name", "Space"),
            NestedTableColumn(
                "vlans",
                "VLANs",
                VlansTable,
                None,
                {
                    "visible_columns": ("vid", "dhcp", "subnets"),
                    "fabrics": fabrics,
                    "subnets": subnets,
                },
            ),
            visible_columns=visible_columns,
        )

    def get_fabric(self, vlan):
        for fabric in self.fabrics:
            for fabric_vlan in fabric.vlans:
                if fabric_vlan.id == vlan.id:
                    return fabric

    def get_vlans(self, vlans):
        for vlan in vlans:
            vlan._data["fabric"] = self.get_fabric(vlan)
        return vlans

    def get_rows(self, target, spaces):
        return (
            (space, self.get_vlans(space.vlans))
            for space in sorted(spaces, key=attrgetter("name"))
        )


class SpaceDetail(DetailTable):
    def __init__(self, *, fabrics=None, subnets=None):
        self.fabrics = fabrics
        super().__init__(
            SpaceNameColumn("name", "Space"),
            NestedTableColumn(
                "vlans",
                "VLANs",
                VlansTable,
                None,
                {
                    "visible_columns": ("vid", "dhcp", "subnets"),
                    "fabrics": fabrics,
                    "subnets": subnets,
                },
            ),
        )

    def get_fabric(self, vlan):
        for fabric in self.fabrics:
            for fabric_vlan in fabric.vlans:
                if fabric_vlan.id == vlan.id:
                    return fabric

    def get_vlans(self, vlans):
        for vlan in vlans:
            vlan._data["fabric"] = self.get_fabric(vlan)
        return vlans

    def get_rows(self, target, space):
        return (space, self.get_vlans(space.vlans))



# Copyright 2016-2017 Canonical Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Asynchronous helpers, for use with `asyncio`."""

__all__ = ["asynchronous", "Asynchronous", "is_loop_running"]

from asyncio import get_event_loop
from functools import wraps
from inspect import isawaitable, iscoroutinefunction


def asynchronous(func):
    """Return `func` in a "smart" asynchronous-aware wrapper.

    If `func` is called within the event-loop — i.e. when it is running — this
    returns the result of `func` without alteration. However, when called from
    outside of the event-loop, and the result is awaitable, the result will be
    passed though the current event-loop's `run_until_complete` method.

    In other words, this automatically blocks when calling an asynchronous
    function from outside of the event-loop, and so makes interactive use of
    these APIs far more intuitive.
    """

    @wraps(func)
    def wrapper(*args, **kwargs):
        eventloop = get_event_loop()
        result = func(*args, **kwargs)
        if not eventloop.is_running():
            while isawaitable(result):
                result = eventloop.run_until_complete(result)
        return result

    return wrapper


class Asynchronous(type):
    """Metaclass that wraps callable attributes with `asynchronous`.

    Use this to create classes instances of which work naturally when working
    with them interactively.
    """

    def __new__(cls, name, bases, attrs):
        attrs.setdefault("__slots__", ())
        attrs = {name: _maybe_wrap(value) for name, value in attrs.items()}
        return super(Asynchronous, cls).__new__(cls, name, bases, attrs)


def _maybe_wrap(attribute):
    """Helper for `Asynchronous`."""
    if iscoroutinefunction(attribute):
        return asynchronous(attribute)
    if isinstance(attribute, (classmethod, staticmethod)):
        if iscoroutinefunction(attribute.__func__):
            return attribute.__class__(asynchronous(attribute.__func__))
    return attribute


def is_loop_running():
    """Is the current event-loop running right now?

    Then we don't want to block, do we?
    """
    return get_event_loop().is_running()



"""Objects for users."""

__all__ = ["User", "Users"]

from . import check, Object, ObjectField, ObjectSet, ObjectType


class UsersType(ObjectType):
    """Metaclass for `Users`."""

    async def whoami(cls):
        """Get the logged-in user."""
        data = await cls._handler.whoami()
        return cls._object(data)

    async def create(cls, username, password, *, email=None, is_admin=False):
        if email is None:
            email = "%s@null.maas.io" % username
        data = await cls._handler.create(
            username=username,
            email=email,
            password=password,
            is_superuser="1" if is_admin else "0",
        )
        return cls._object(data)

    async def read(cls):
        data = await cls._handler.read()
        return cls(map(cls._object, data))


class Users(ObjectSet, metaclass=UsersType):
    """The set of users."""


class UserType(ObjectType):
    async def read(cls, username):
        data = await cls._handler.read(username=username)
        return cls(data)


class User(Object, metaclass=UserType):
    """A user."""

    username = ObjectField.Checked("username", check(str), check(str), pk=True)
    email = ObjectField.Checked("email", check(str), check(str))
    is_admin = ObjectField.Checked("is_superuser", check(bool), check(bool))

    def __repr__(self):
        if self.loaded and self.is_admin:
            return super(User, self).__repr__(name="Admin", fields={"username"})
        else:
            return super(User, self).__repr__(name="User", fields={"username"})
